package uk.co.twey.uni.dooom;

/**
 *  Represents an action to be carried out on receipt of a command.
 *  Responsible for validation of the command and also execution.
 */
public abstract class CommandOfDooom {
    // We only have nullary and unary commands in the spec, & the
    // parsing algorithm assumes this, but just in case…
    protected int argc = -1;
    protected int ap;
    protected boolean requiresPlayer;

    /**
     * Create a new command with the default values (variadic
     * arguments, no action points required, doesn't require the
     * client to be playing).
     */
    public CommandOfDooom() {
        this(-1, 0, false);
    }

    /**
     * Create a new command.
     *
     * @param argc the number of arguments the command should take.
     * -1 indicates variadic arguments.
     *
     * @param ap the number of action points required to execute this
     * command; if greater than zero, a player is also required.
     */
    public CommandOfDooom(int argc, int ap) {
        this.argc = argc;
        this.ap = ap;
        this.requiresPlayer = ap > 0;
    }

    /**
     * Create a new command.
     *
     * @param argc the number of arguments the command should take.
     * -1 indicates variadic arguments.
     * @param ap the number of action points required to execute this
     * command.
     * @param requiresPlayer whether the client must be playing and
     * attached to a {@link PlayerOfDooom} object in order to use this
     * command.
     */
    public CommandOfDooom(int argc, int ap, boolean requiresPlayer) {
        this.argc = argc;
        this.ap = ap;
        this.requiresPlayer = requiresPlayer;
    }

    /**
     * Execute the command.  This should be overridden by individual
     * commands.
     *
     * @param client the {@link ClientOfDooom} from which the command
     * originated.
     * @param cmd the command executed.
     * @param args the arguments passed to the command; length is
     * guaranteed to equal to the argc specified in the constructor,
     * with the exception of -1.
     */
    public abstract void execute(ClientOfDooom client, String cmd,
                                 String... args);

    /**
     * Get the usage example to be displayed when the user attempts to
     * use a command incorrectly.
     */
    protected abstract String getUsage();

    /**
     * Print out the usage of the command to the client.
     *
     * @param client the client to inform.
     */
    public void printUsage(ClientOfDooom client) {
        try {
            client.writeMessage("Usage: " + getUsage());
        } catch (java.io.IOException e) {
            System.err.println("Failed to write usage:\n" + e);
        }
    }

    /**
     * Ensure that the command given meets the requirements of this
     * Command object; if so, execute it.  If not, inform the client
     * of the error.
     *
     * @param client the client attempting to execute the command.
     * @param cmd the command being executed.
     * @param args the arguments passed to the command.
     */
    public void checkedExecute(ClientOfDooom client, String cmd,
                               String... args) {
        if (argc >= 0 && args.length != argc) {
            printUsage(client);
            return;
        }

        if (client.getPlayer() == null && requiresPlayer)
            try {
                client.writeMessage("FAIL You must have a player to execute"
                                    + " this command.");
                return;
            } catch (java.io.IOException e) {
                System.err.println("Failed to print error:\n" + e);
                System.exit(4);
            }

        if (ap > client.getPlayer().getActionPoints()) {
            client.getPlayer().tell("FAIL You don't have enough action points ("
                                    + ap + ") to execute this command!");
            return;
        }

        if (requiresPlayer)
            client.getPlayer().subtractActionPoints(ap);

        execute(client, cmd, args);
    }
}