package uk.co.twey.uni.dooom;

import java.util.*;

/**
 * Represents a single game played on a particular map.
 */
public class GameOfDooom extends CommandProcessorOfDooom {
    /** The map on which the game is being played. */
    protected MapOfDooom map;
    public MapOfDooom getMap() { return map; }

    /** The players playing the game. */
    protected List<PlayerOfDooom> players = new LinkedList<PlayerOfDooom>();

    /** The current turn. */
    protected int turn = -1;
    public int getTurn() { return turn; }

    /**
     * Get the current player.
     *
     * @return the current player.
     */
    public PlayerOfDooom getCurrentPlayer() {
        return (turn < 0 || turn >= players.size()) ? null : players.get(turn);
    }

    /**
     * Switch to the next turn.
     */
    public void endTurn() {
        PlayerOfDooom player = getCurrentPlayer();

        if (player != null) {
            player.tell("ENDTURN");
            player.setActionPoints(0);
        }

        if (++turn >= players.size()) turn = 0;

        player = getCurrentPlayer();
        player.resetActionPoints();
        player.tell("STARTTURN");
    }

    /**
     * Create a new game.
     *
     * @param map the map on which the game is being played.
     */
    public GameOfDooom(MapOfDooom map) {
        this.map = map;

        final GameOfDooom game = this;

        commands = new java.util.HashMap<String, CommandOfDooom>();

        commands.put("LOOK", new CommandOfDooom(0, 0, true) {
                protected String getUsage() { return "LOOK"; }
                
                public void execute(ClientOfDooom client, String cmd,
                                    String... args) {
                    try {
                        client.writeMessage("LOOKREPLY\n" + game.map.look(
                            (FloorOfDooom) client.getPlayer().getHolder(),
                            client.getPlayer().getVision()));
                    } catch (java.io.IOException e) {
                        System.err.println("Failed to print map:\n" + e);
                    }
                }
            });

        commands.put("HELLO", new CommandOfDooom(-1, 0, false) {
                protected String getUsage() { return "HELLO [name]"; }

                public void execute(ClientOfDooom client, String cmd,
                                    String... args) {

                    if (args.length > 0)
                        client.getPlayer().setPlayerName(args[0]);

                    try {
                        client.writeMessage("GOLD " + game.map.getGoal());
                    } catch (java.io.IOException e) {
                        System.err.println("Failed to respond to HELLO:\n" + e);
                    }
                }
            });

        commands.put("PICKUP", new CommandOfDooom(0, 0, true) {
                protected String getUsage() { return "PICKUP"; }

                public void execute(ClientOfDooom client, String cmd,
                                    String... args) {
                    PlayerOfDooom player = client.getPlayer();
                    LinkedList<ObjectOfDooom> toBeTaken
                        = new LinkedList<ObjectOfDooom>();
                    boolean picked = false;

                    // Somewhat awkward algorithm because Java doesn't like us
                    // modifying collections we're iterating over.

                    for (ObjectOfDooom thing : player.getHolder().getItems())
                        toBeTaken.add(thing);

                    for (ObjectOfDooom thing : toBeTaken)
                        try {
                            thing.giveTo(player);
                            picked = true;
                        } catch (PossessionException e) { }

                    if (picked)
                        player.tell("SUCCESS, GOLD COINS: " + player.getGold());
                    else
                        player.tell("FAIL");
                }
            });

        commands.put("MOVE", new CommandOfDooom(1, 1, true) {
                protected String getUsage() { return "MOVE [NESW]"; }

                public void execute(ClientOfDooom client, String cmd,
                                    String... args) {
                    PlayerOfDooom player = client.getPlayer();

                    PlaceOfDooom newPlace,
                        oldPlace = (PlaceOfDooom) player.getHolder();

                    try {
                        switch (args[0].charAt(0)) {
                        case 'N':
                            player.move(game.getMap(), 0, -1); break;
                        case 'E':
                            player.move(game.getMap(), 1, 0); break;
                        case 'S':
                            player.move(game.getMap(), 0, 1); break;
                        case 'W':
                            player.move(game.getMap(), -1, 0); break;
                        default:
                            printUsage(client);
                            return;
                        }

                    } catch (ImpassableException e) {
                        player.tell("FAIL");
                        return;
                    } catch (PlaceOccupiedException e) {
                        player.tell("FAIL This space is occupied by "
                                    + e.getBlocker().getName() + ".");
                        return;
                    } catch (ExitException e) {
                        player.tell("SUCCESS");

                        tryWin(player);
                    }

                    newPlace = (PlaceOfDooom) player.getHolder();

                    Set<PlayerOfDooom> observers = getObservers(oldPlace);
                    observers.addAll(getObservers(newPlace));

                    for (PlayerOfDooom p : observers)
                        p.tell("CHANGE");

                    player.tell("SUCCESS");
                }
            });

        commands.put("ENDTURN", new CommandOfDooom(0, 0, true) {
                protected String getUsage() { return "ENDTURN"; };

                public void execute(ClientOfDooom client, String cmd,
                                    String... args) {
                    endTurn();
                }
            });
    }

    /**
     * See if a player has enough gold to win; if they do, they win!
     * Exit the game if so.
     *
     * @param winner the player who is trying to win.
     */
    protected void tryWin(PlayerOfDooom winner) {
        if (winner.getGold() >= getMap().getGoal()) {
            for (PlayerOfDooom player : players)
                if (player.equals(winner))
                    player.tell("WIN");
                else
                    player.tell("LOSE");

            System.exit(0);
        }
    }

    protected void onUnknown(ClientOfDooom client, String cmd,
                             String... arg) {
        try {
            client.writeMessage("Unknown command: " + cmd);
        } catch (java.io.IOException e) {
            System.out.println("Failed to respond to unknown message \""
                               + cmd + "\":\n" + e);
        }
    }

    /**
     * Send a message to all players in the game.
     *
     * @param message the message to send.
     */
    public void broadcast(String message) {
        for (PlayerOfDooom player : players)
            player.tell(message);
    }

    /**
     * Add a player to the game.
     *
     * @param player the player to add.
     *
     * @return whether the player could be added.
     */
    public boolean addPlayer(PlayerOfDooom player) {
        if (players.contains(player))
            return false;

        players.add(player);

        if (players.size() == 1) {
            turn = -1;
            endTurn();
        }

        boolean r;

        try {
            r = player.placeOn(map);
        } catch (ExitException e) {
            tryWin(player);
            r = true;
        }

        if (!r) return false;

        for (PlayerOfDooom p : getObservers((PlaceOfDooom) player.getHolder()))
            p.tell("CHANGE");

        return true;
    }

    /**
     * Get observing players.
     *
     * @param place the place being observed.
     *
     * @return a set of observing players.
     */
    public Set<PlayerOfDooom> getObservers(PlaceOfDooom place) {
        Set<PlayerOfDooom> r = new HashSet<PlayerOfDooom>();

        for (PlayerOfDooom p : players)
            if (Math.round(((PlaceOfDooom) p.getHolder()).distanceFrom(place))
                  <= p.getVision())
                r.add(p);

        return r;
    }

    /**
     * Remove a player from the game.
     *
     * @param player the player to remove.
     */
    public void removePlayer(PlayerOfDooom player) {
        if (getCurrentPlayer().equals(player))
            endTurn();

        player.die();

        if (players.indexOf(player) < turn)
            --turn;

        players.remove(player);
    }
}