package uk.co.twey.uni.dooom;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.List;
import java.util.LinkedList;

/** Class representing a map. */
public class MapOfDooom {
    /** The map, represented as a two-dimensional array. */
    protected PlaceOfDooom[][] places;

    /** The name of the map, given by the "name" field. */
    protected String name;
    public String getName() { return name; }

    /** The amount of gold required to win this map, given by the
     * "win" field. */
    protected int goal;
    public int getGoal() { return goal; }

    /** The width of the map. */
    protected int width;
    public int getWidth() { return width; }

    /** The height of the map. */
    protected int height;
    public int getHeight() { return height; }

    /**
     * Convenience function to return null rather than throw an
     * IndexOutOfBoundsException and to reverse the order of x and y
     * into their common usages.
     *
     * @param x the x position of the place to get.
     * @param y the y position of the place to get.
     *
     * @return the place at these coördinates.
     */
    public PlaceOfDooom getPlace(int x, int y) {
        try {
            return places[y][x];
        } catch (IndexOutOfBoundsException e) {
            return null;
        }
    }

    /**
     * Convenience function to update the contents of a place, à la
     * setPlace().
     */
    protected void setPlace(int x, int y, PlaceOfDooom place) {
        places[y][x] = place;
    }

    /**
     * Create a new MapOfDooom.
     *
     * @param name the title of the map, given by the "name" field.
     * @param goal the amount of gold needed to win the map, given by the "win"
     * field.
     */
    public MapOfDooom(String name, int goal, String map)
            throws MapFormatException {
        this.name = name;
        this.goal = goal;

        List<String> lines
            = new LinkedList<String>(Arrays.asList(map.split("\n")));
        lines.removeAll(Collections.singleton(""));

        if (lines.size() == 0)
            throw new MapFormatException("Empty map");

        height = lines.size();
        width = lines.get(0).length();
        places = new PlaceOfDooom[height][width];

        int y = 0;
        for (String line : lines) {
            if (line.length() != width)
                throw new MapFormatException("Non-rectangular map");

            for (int x = 0, nx = line.length(); x < nx; ++x)
                switch (line.charAt(x)) {
                case '#':
                    setPlace(x, y, new WallOfDooom(x, y));
                    break;
                case 'E':
                    setPlace(x, y, new ExitOfDooom(x, y));
                    break;
                case '.':
                    setPlace(x, y, new FloorOfDooom(x, y));
                    break;
                case 'G':
                    setPlace(x, y, new FloorOfDooom(x, y, new GoldOfDooom()));
                    break;
                case ' ':
                    setPlace(x, y, new DarknessOfDooom(x, y));
                    break;
                default:
                    throw new MapFormatException("Unknown symbol '"
                                                 + line.charAt(x) + "'");
                }

            ++y;
        }
    }

    /**
     * Load a map from a file.
     *
     * @param file the file from which to load the map.
     *
     * @throws IOException if an error occurred reading the file.
     * @throws MapFormatException if the map format is incorrect.
     *
     * @return the map successfully loaded from the file.
     */
    public static MapOfDooom fromFile(File file)
            throws IOException, MapFormatException {
        int goal = -15;
        String name = null, line;
        List<String> map = new LinkedList<String>();
        BufferedReader in = new BufferedReader(new FileReader(file));

        while ((line = in.readLine()) != null)
            if (line.startsWith("name "))
                name = line.substring(5);
            else if (line.startsWith("win "))
                try {
                    goal = Integer.parseInt(line.substring(4));
                } catch (NumberFormatException e) {
                    throw new MapFormatException("Non-integer goal");
                }
            else
                map.add(line);

        if (goal == -15)
            throw new MapFormatException("No goal specified");
        if (goal < 0)
            throw new MapFormatException("Negative goal");
        if (name == null)
            throw new MapFormatException("No name specified");
        
        return new MapOfDooom(name, goal, Util.join(map, "\n"));
    }

    /**
     * Get all free floor available on this map.
     *
     * @return the set of all the free floor places in the map.
     */
    public Set<FloorOfDooom> getFloor() {
        HashSet<FloorOfDooom> r = new HashSet<FloorOfDooom>();

        for (PlaceOfDooom[] row : places)
            place: for (PlaceOfDooom place : row)
                if (place instanceof FloorOfDooom) {
                    for (ObjectOfDooom item : ((FloorOfDooom) place).getItems())
                        if (item instanceof PlayerOfDooom)
                            continue place;

                    r.add((FloorOfDooom) place);
                }

        return r;
    }

    /**
     * Get a string representation of the map.
     *
     * @return the map represented in the string format used in the map files.
     */
    public String getPlan() {
        StringBuilder r = new StringBuilder();

        for (int y = 0; y < places.length; ++y) {
            for (int x = 0; x < places[y].length; ++x)
                r.append(places[y][x].getIcon());
            r.append('\n');
        }
       
        return r.toString();
    }

    /**
     * Display one square in each direction from a given position.
     *
     * @param origin the starting point around which to view terrain.
     *
     * @return symbols representing the contents of the terrain,
     * arranged in NSEW order.
     */
    public String look(PlaceOfDooom origin, int distance) {
        char[][] out = new char[2 * distance + 1][2 * distance + 1];
        int x = origin.getX(), y = origin.getY();

        out[distance][distance] = ((FloorOfDooom) origin).getIconNoPlayer();

        for (int yd = -distance; yd <= distance; ++yd)
            for (int xd = -distance; xd <= distance; ++xd)
                if (xd == 0 && yd == 0)
                    out[distance][distance] = ((FloorOfDooom) origin).getIconNoPlayer();
                else {
                    PlaceOfDooom place = getPlace(x + xd, y + yd);
                    out[yd + distance][xd + distance] =
                        place == null ? 'X'
                        : Math.round(place.distanceFrom(origin)) <= distance ? place.getIcon()
                        : 'X';
                }

        StringBuilder sb = new StringBuilder();

        for (int i = 0; i < out.length; ++i) {
            sb.append(out[i]);
            if (i < out.length - 1)
                sb.append('\n');
        }

        return sb.toString();
    }

    /**
     * Helper method
     */
    private static float dist(float x1, float y1, float x2, float y2) {
        return (float) Math.sqrt(Math.pow(x2 - x1, 2) + Math.pow(y2 - y1, 2));
    }

    /**
     * {@inheritDoc}
     * @return the map in map-file format.
     */
    public String toString() {
        return "name " + getName() + "\n"
            + "win " + getGoal() + "\n"
            + getPlan();
    }
}