package uk.co.twey.uni.dooom;

import java.util.*;

/** A class representing a player; that is, the avatar of a human or
 * AI on the map, as distinct from the {@link ClientOfDooom} which
 * represents the real player and handles interfacing with them. */
public class PlayerOfDooom extends ObjectOfDooom implements HolderOfDooom {
    /** The name by which this player is to be referred in-game. */
    protected String playerName;
    public String getPlayerName() { return playerName; }
    public void setPlayerName(String name) { playerName = name; }

    /** The amount of gold currently carried by this player. */
    protected int gold = 0;
    public int getGold() { return gold; }
    /** Add an amount of gold to the player's current total. */
    public void addGold(int amount) { gold += amount; }

    /** The number of action points remaining to this player.  When
     * the player runs out of action points, their turn ends. */
    protected int actionPoints = 0;
    public int getActionPoints() { return actionPoints; }
    public void setActionPoints(int ap) { actionPoints = ap; }
    /** Subtract a number of action points from the player's current
     * total. */
    public void subtractActionPoints(int amount) { actionPoints -= amount; }

    /** The maximum action points this player may have, and the number
     * granted at the start of each turn. */
    protected int maxActionPoints = 6;
    public int getMaxActionPoints() { return maxActionPoints; }

    /** Not currently used. */
    protected int strength = 1;
    public int getStrength() { return strength; }

    /** Not currently used. */
    protected int defence = 0;
    public int getDefence() { return defence; }

    /** The distance the player can see. */
    protected int vision = 2;
    public int getVision() { return vision; }

    /** The icon used to represent this player on the map. */
    protected char icon = 'P';
    public char getIcon() { return icon; }
    public int getPrecedence() { return 10; }
    public String getName() { return getPlayerName(); }

    /** The client with which this player is associated. */
    protected ClientOfDooom client;
    public ClientOfDooom getClient() { return client; };

    protected Set<ObjectOfDooom> items = new HashSet<ObjectOfDooom>();
    public Set<ObjectOfDooom> getItems() { return items; }

    /**
     * Create a new PlayerOfDooom with the default icon of P.
     *
     * @param playerName the name by which this player is to be
     * addressed in-game.
     * @param client the client associated with this player.
     *
     * @throws IllegalArgumentException if a client is passed that
     * already has a player associated with it.
     */
    public PlayerOfDooom(String playerName, ClientOfDooom client) {
        this.playerName = playerName;
        this.client = client;

        if (client.getPlayer() != null)
            throw new IllegalArgumentException("Client already has a player ("
                + client.getPlayer().getPlayerName() + ") attached");
        else
            client.setPlayer(this);
    }

    /**
     * Create a new PlayerOfDooom.
     *
     * @param playerName the name by which this player is to be
     * addressed in-game.
     * @param client the client associated with this player.
     * @param icon the icon to be used to represent this player on the map.
     *
     * @throws IllegalArgumentException if a client is passed that
     * already has a player associated with it.
     */
    public PlayerOfDooom(String playerName, ClientOfDooom client, char icon) {
        this(playerName, client);

        this.icon = icon;
    }

    /**
     * Attempt to send a message to this player's client, ignoring any
     * errors that may occur.
     *
     * @param message the message to send.
     */
    public void tell(String message) {
        try {
            getClient().writeMessage(message);
        } catch (java.io.IOException e) {
            System.err.println("Failure to communicate with client " + this + ":\n" + e);
            System.exit(3);
        }
    }

    /**
     * Reset the player's action points to their maximum.
     */
    public void resetActionPoints() {
        actionPoints = maxActionPoints;
    }

    public void take(ObjectOfDooom item) {
        items.add(item);
    }

    public void drop(ObjectOfDooom item) {
        items.remove(item);
    }

    /**
     * {@inheritDoc}
     * <p>
     * Players may only be held by {@link FloorOfDooom} and its subclasses.
     */
    public void onTake(HolderOfDooom holder) throws PossessionException {
        if (!(holder instanceof FloorOfDooom))
            throw new PossessionException(this);
    }

    /**
     * Randomly place the player on a floor space on the given map.
     *
     * @param map the map on which to place the player.
     *
     * @return false if the player could not be placed, else true
     */
    public boolean placeOn(MapOfDooom map) throws ExitException {
        FloorOfDooom[] floor = map.getFloor().toArray(new FloorOfDooom[] { });
        int start = (new Random()).nextInt(floor.length);
        int next = start;

        while (true)
            try {
                FloorOfDooom target = floor[next];
                giveTo(target);
                target.onEnter(this);
                return true;
            } catch (PossessionException e) {
                System.err.println(e.toString()
                                   + "\nCan't give player to floor; "
                                   + "can't happen!  Bailing out…");
                System.exit(2);
            } catch (ExitException e) {
                throw e;
            } catch (EntryException e) {
                if (++next >= floor.length)
                    next = 0;

                if (next == start)
                    return false;
            }
    }

    /**
     * Move the player a set number of places along the x and y axes.
     *
     * @param map the map on which to move the player.
     * @param xVector the number of squares rightwards to move the player.
     * @param yVector the number of squares downwards to move the player.
     *
     * @throws ImpassableException if the move would result in the
     * player landing on impassable terrain (such as a wall).
     * @throws ExitException if the player lands on an exit.
     */
    public void move(MapOfDooom map, int xVector, int yVector)
        throws ImpassableException, PlaceOccupiedException, ExitException {
        FloorOfDooom source = (FloorOfDooom) getHolder();
        PlaceOfDooom target = map.getPlace(source.getX() + xVector,
                                           source.getY() + yVector);

        if (target == null || !target.isPassable())
            throw new ImpassableException(this, target);

        try {
            try {
                target.onEnter(this);
                giveTo((FloorOfDooom) target);
            } catch (PlaceOccupiedException e) {
                throw e;
            } catch (ExitException e) {
                giveTo((FloorOfDooom) target);
                throw e;
            } catch (EntryException e) {
                System.err.println("Unknown exception occurred as a result of "
                                   + "entering " + target + ":\n" + e);
                System.exit(6);
            }
        } catch (PossessionException e) {
            System.err.println("Failed to give player to floor: shouldn't "
                               + "happen!  Bailing out…\n" + e);
            System.exit(5);
        }
    }

    /**
     * Make the player die, dropping all applicable items.
     * Does not remove player from player list!
     */
    synchronized public void die() {
        HolderOfDooom floor = getHolder();

        floor.drop(this);

        for (ObjectOfDooom item : getItems())
            try {
                item.giveTo(floor);
            } catch (PossessionException e) {
            }
    }
}
