package uk.co.twey.uni.dooom.client;

import uk.co.twey.uni.dooom.*;

import java.io.*;
import javax.swing.*;

/**
 * Represents, constructs, and provides a handle to the client UI for
 * other aspects of the client to use.
 */
public class ClientUI implements Runnable {
    /** The outward-facing connection to the server. */
    protected PrintWriter out;
    /** The main window. */
    protected ClientUIWindow win;
    /** Whether the game has ended. */
    protected boolean ended = false;

    /**
     * Create a new ClientUI.
     *
     * @param out the outward-facing connection to the server.
     * User-initiated commands will be written here.
     */
    public ClientUI(PrintWriter out) {
        this.out = out;

        try {
            UIManager.setLookAndFeel(
                UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {
            System.err.println("Couldn't set L&F:\n" + e);
            // Never mind then
        }
            
        SwingUtilities.invokeLater(this);
    }

    public void run() {
        win = new ClientUIWindow(out);
    }

    /**
     * Receive and process a command line.
     *
     * @param line the line to process.
     * @param in where to get more follow-up lines, if necessary.
     */
    public void receiveLine(String line, BufferedReader in) {
        appendLog(">> " + line);

        if ("LOOKREPLY".equals(line))
            try {
                StringBuilder lines = new StringBuilder();

                line = in.readLine();

                lines.append(line);

                for (int i = 0, n = line.length() - 1; i < n; ++i)
                    lines.append("\n" + in.readLine());

                updateMap(lines.toString());
            } catch (IOException e) {
                System.err.println("ClientUI.receiveLine(): "
                                   + "LOOKREPLY terminated prematurely.");
            }
        else if ("CHANGE".equals(line)) {
            appendLog("<< LOOK");
            out.println("LOOK");
        } else if ("WIN".equals(line))
            SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        ended = true;
                        JOptionPane.showMessageDialog(
                            win, "You won.", "Game Result",
                            JOptionPane.INFORMATION_MESSAGE);
                        System.exit(0);
                    }
                });
        else if ("LOSE".equals(line))
            SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        ended = true;
                        JOptionPane.showMessageDialog(
                            win, "You lose.", "Game Result",
                            JOptionPane.INFORMATION_MESSAGE);
                        System.exit(0);
                    }
                });
        else if (line == null)
            SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        if (!ended) {
                            JOptionPane.showMessageDialog(
                              win, "Server disconnected.", "Error",
                              JOptionPane.ERROR_MESSAGE);
                            System.exit(0);
                        }
                    }
                });
        else if (line.startsWith("GOLD "))
            try {
                final int goal = Integer.parseInt(line.substring(5));
                SwingUtilities.invokeLater(new Runnable() {
                        public void run() {
                            win.setGoal(goal);
                        }
                    });
            } catch (NumberFormatException e) { }
        else if (line.startsWith("SUCCESS, GOLD COINS: "))
            try {
                final int gold = Integer.parseInt(line.substring(
                                    "SUCCESS, GOLD COINS: ".length()));
                SwingUtilities.invokeLater(new Runnable() {
                        public void run() {
                            win.setGold(gold);
                        }
                    });
            } catch (NumberFormatException e) { }
    }

    /**
     * Update the map in response to a LOOKREPLY.
     *
     * @param plan the map data.
     */
    public void updateMap(final String plan) {
        SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    win.updateMap(plan);
                }
            });
    }

    /**
     * Append a line to the log.
     *
     * @param msg the message to add.
     */
    public void appendLog(final String msg) {
        SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    win.appendLog(msg);
                }
            });
    }
}