package uk.co.twey.uni.dooom.server;

import uk.co.twey.uni.dooom.*;

import java.io.*;
import java.net.*;

/**
 * Represents a client connected to the server.
 */
public class ServerClient extends ClientOfDooom implements Runnable {
    /** The number of clients seen so far. */
    static int count = 0;
    /** Increment the count. */
    static synchronized int incCount() { return count++; }

    /** The socket to which the client is connected. */
    protected Socket sock;
    /** The input stream from the client. */
    protected BufferedReader in;
    /** The output stream to the client. */
    protected PrintWriter out;
    /** The {@link Server} instance to which the client is connected. */
    protected Server server;
    /** The {@link ServerUI} controlling the server. */
    protected ServerUI ui;

    /** The {@link PlayerOfDooom} controlled via this client. */
    protected PlayerOfDooom player;
    public PlayerOfDooom getPlayer() { return player; }
    public void setPlayer(PlayerOfDooom player) { this.player = player; }

    /**
     * Create a new ServerClient.
     *
     * @param server the {@link Server} object to which this client is
     * connected.
     * @param ui the {@link ServerUI} controlling said server.
     * @param sock the socket to which the client is connected.
     */
    public ServerClient(Server server, ServerUI ui, Socket sock) {
        this.server = server;
        this.sock = sock;
        this.ui = ui;
        try {
            in = new BufferedReader(
                   new InputStreamReader(sock.getInputStream()));
            out = new PrintWriter(sock.getOutputStream(), true);
        } catch (IOException e) {
            System.err.println("Could not communicate with client:\n" + e);
            System.exit(8);
        }
        this.player = new PlayerOfDooom("Player " + ServerClient.incCount(), this);
        ui.appendLog(" >> " + player.getPlayerName() + " joined <<");
    }

    /**
     * Process the client.
     */
    public void run() {
        try {
            synchronized (server) {
                server.getGame().addPlayer(player);
                ui.updateMap();
            }

            for (String cmd; (cmd = getCommand()) != null; )
                synchronized (server) {
                    server.getGame().receiveLine(this, cmd);
                    ui.updateMap();
                }
        } finally {
            try {
                in.close();
                out.close();
                sock.close();
            } catch (IOException e) { }

            synchronized (server) {
                server.getGame().removePlayer(player);
                ui.updateMap();
            }
        }
    }

    /**
     * Read a command from the client.
     *
     * @return the new command, or null if the client has disconnected.
     */
    public String getCommand() {
        try {
            String cmd = in.readLine();
            if (cmd == null)
                ui.appendLog(" >> " + player.getPlayerName() + " dropped <<");
            else
                ui.appendLog(player.getPlayerName() + " >>\n" + cmd);
            return cmd;
        } catch (IOException e) {
            return null;
        }
    }

    /**
     * Write a message to the client.
     *
     * @param msg the message to write.
     */
    public void writeMessage(String msg) {
        ui.appendLog(getPlayer().getPlayerName() + " <<\n" + msg);
        out.println(msg);
    }
}